/*
 * author: bto
 * date: 20 jun 2020
 * 4-bit HD44780U LCD driver
 */

#include "lcd_driver.h"
#include "olimex_avr-p20.h"
#include "useful.h"
#include "uart.h"
#include "util/delay.h"

/* For this code to function, the following must be defined:
 *
 * LCD_BUS = lcd data register
 * D4,D5,D6,D7 = data pins
 * LCD_E  = lcd enable
 * LCD_RW = lcd read/write
 * LCD_RS = lcd register select
 */

// definitions for a HD44870U LCD

#define LCD_ENABLE  setbit(LCD_REG,LCD_E)
#define LCD_DISABLE clrbit(LCD_REG,LCD_E)
#define LCD_READ    setbit(LCD_REG,LCD_RW)
#define LCD_WRITE   clrbit(LCD_REG,LCD_RW)
#define LCD_STR     setbit(LCD_REG,LCD_RS)
#define LCD_INS     clrbit(LCD_REG,LCD_RS)

unsigned char i; // counter

lcd_tx(unsigned char d)
{
    // by design, this function does not know whether d is a character to
    // be displayed or data to be input, the mode is defined in the calling 
    // function. technically, the delays between commands are different
    // between read/write, but they're all smaller than 1us anyway.
    
    LCD_BUS &= 0xF0; // mask old nibble 
    LCD_ENABLE;
    LCD_BUS |= HI_NIBBLE(d); // send high nibble
    _delay_us(1);
    LCD_DISABLE;

    LCD_BUS &= 0xF0; // mask old nibble
    LCD_ENABLE;
    LCD_BUS |= LO_NIBBLE(d); // send low nibble
    _delay_us(1);
    LCD_DISABLE;
}

void
lcd_print(unsigned char d)
{
    // print a character
    LCD_STR;
    LCD_WRITE;
    LCD_DISABLE;
    lcd_tx(d);
    _delay_us(40);
}

void
lcd_cmd(unsigned char d)
{
    // send the LCD a command
    LCD_INS;
    LCD_WRITE;
    LCD_DISABLE;
    lcd_tx(d);
    _delay_us(40);
}

void
shift_cursor_left(unsigned char n)
{
    for (i=0; i<n; i++) lcd_cmd(0x10); // 0b00010000 shift cursor left
}

void
shift_cursor_right(unsigned char n)
{
    for (i=0; i<n; i++) lcd_cmd(0x14); // 0b00010100 shift cursor right
}

void
shift_display_left(unsigned char n)
{
    for (i=0; i<n; i++) lcd_cmd(0x18); // 0b00011000 shift display left
}

void
shift_display_right(unsigned char n)
{
    for (i=0; i<n; i++) lcd_cmd(0x1C); // 0b00011100 shift display right
}

void 
lcd_putc(char c) // write char to lcd
{
    lcd_print(c);
}

void
lcd_puts(const char *s) // write string to lcd
{
    while (*s) {
        lcd_print(*s++);
    }
}

void
lcd_goto(unsigned char r, unsigned char c)
{
    // LCD wraps line 0->2 and 1->3
    lcd_home();
    if (r == 0) shift_cursor_right(c);
    if (r == 1) shift_cursor_right(40 + c);
    if (r == 2) shift_cursor_right(20 + c);
    if (r == 3) shift_cursor_right(60 + c);
}

void
lcd_clear()
{
    lcd_cmd(0x01); // 0b00000001, clear display, 1.53ms
    _delay_ms(2);
}

void
lcd_home(void)
{
    lcd_cmd(0x02); // 0b00000010, home cursor, 1.53ms
    _delay_ms(2);
}

void
setup_lcd(void)
{
    _delay_ms(50);

    LCD_INS; 
    LCD_WRITE;
    LCD_ENABLE;
    
    // put lcd in 4-bit mode, while the lcd starts in 8-bit mode.
    // basically, send a byte that sets D5 = 1, D4 = 0.
    lcd_tx(0x03); _delay_ms(10); // 0b00000011

    // p.13, initialise 4-bit mode
    lcd_tx(0x30); _delay_us(100); // 0b00110000, function set
    lcd_tx(0x30); _delay_us(100); // 0b00110000, function set
    lcd_tx(0x20); _delay_us(40);  // 0b00100000, function set: 4-bit mode
    lcd_tx(0x28); _delay_us(40);  // 0b00101000, function set: 4-bit mode, 2-line, 5x10 dots, 39us
    lcd_tx(0x08); _delay_ms(2);   // 0b00001000, display off
    lcd_tx(0x01); _delay_ms(2);   // 0b00000001, display clear, 1.53ms
    lcd_tx(0x06); _delay_us(40);  // 0b00000110, entry mode set: cursor dir right, display shift off, 39us
    //lcd_tx(0x0F); _delay_us(40);  // 0b00001111, display on, cursor on, blink on, 39us
    lcd_tx(0x0C); _delay_us(40);  // 0b00001100, display on, cursor off, blink off, 39us
}
