// author: bto
// date: 11 jan 2015
// serial communication program for AT90S2313

#include "olimex_avr-p20.h"
#include "useful.h"
#include "uart.h"
#include "avr/interrupt.h"

// For 9-bit transmission, the 9th bit indicates address (1) or data (0).
// This allows subsequent data packets to be ignored if required.

/* for this code to function, the following must be defined:
 * 
 * BAUD
 *
 */

ISR(UART_TX_vect)
{
    /* 
     * UART Tx process
     * 1. char shifted out from transmit shift register
     * 2. no new data in UDR (TXC is set)
     * 3. if TXCIE, interrupt called
     * 4. TXC cleared manually
     *
     * Therefore, this interrupt is called every time a char is transmitted
     * Useful in duplex comm when app must go into Rx mode immediately after Tx
     */
    
     setbit(USR,TXC); // clear bit
}

ISR(UART_RX_vect)
{
    /*
     * UART Rx process
     * 1. char arrives in receive shift register
     * 2. char transferred to UDR (RXC is set)
     * 3. if RXCIE, interrupt called
     * 4. RXC cleared by reading UDR
     * 
     * Therefore, this interrupt is called every time a char arrives
     */

     if (readbit(USR,FE)) {} // check framing error before reading UDR
     char c;
     uart_getc(&c); // receive char
     uart_putc(c);  // echo char
     if (readbit(USR,DOR)) {} // check overrun error after reading UDR
}

// convert binary character to ASCII
unsigned char
bin2hex(signed char bin)
{
    if ((0 <= bin) && (bin <= 9)) return ((unsigned)bin + '0');
    if ((10 <= bin) && (bin <= 15)) return ((unsigned)bin - 10 + 'A');
    return '?';
}

// split an int into an array of chars  
void
split_int(unsigned int *val, unsigned char *int_data)
{
    int_data[3] = (unsigned char)(*val/1000);
    int_data[2] = (unsigned char)((*val%1000)/100);
    int_data[1] = (unsigned char)((*val%100)/10);
    int_data[0] = (unsigned char)(*val%10);
}

/*
char*
int2str(unsigned int *src, unsigned char *dest, unsigned char len)
{
    for 
}

char *
strcpy(char *dest, const char *src)
{
    char *save = dest;
    while (*dest++ = *src++);
    return save;
}
*/
void
uart_getc(char *c)
{
    while(!readbit(USR,RXC)); // wait until receive register is empty
    *c = UDR; // get character
}

void
uart_putc(const char c)
{
    // replace newline with carriage return
    if (c == '\n') {
        uart_putc('\r');
    }
    while (!readbit(USR,UDRE)); // wait until data register is empty
    UDR = c; // load data register
}

void
uart_puts(const char *s)
{
    while (*s) { // while there is data in the string
        uart_putc(*s++); // put char in UDR
    }
}

void
setup_uart(void)
{
    // set baudrate (p.47)
    UBRR = F_CPU/(16L*BAUD) - 1;
    
    // status register
    SREG |= 0x80; // 0b10000000 enable global interrupts

    // control register
    UCR  |= 0x98; // 0b10011000 enable interrupts, tx, rx (p.46)
    //UCR  |= 0x18; // 0b00011000 enable tx, rx only (p.46)
    //UCR  |= 0x00; // 0b00000000 set PD0 and PD1 as GPIO

    // status register
    USR |= 0x40; // 0b01000000 clear interrupt flags
}

/*
int
setup(void)
{
    DDRB  = 0x80; // 0b10000000 set PORTB pin 7 as output
    LED0_OFF;

    // data registers
    DDRD  = 0x01; // 0b00000001 set Rx (PD0) as input, Tx (PD1) as output
    //DDRD  = 0x00; // 0b00000000 test: set PD0 and PD1 as output
    PORTD = 0x00; // 0b00000000 clear values
    
    return SUCCESS;
}

int 
main(void)
{
    setup();
    char tx_data[] = "hello, world!";
    
    while (TRUE) {
        uart_puts(tx_data);
        _delay_ms(1000);
    }

    return SUCCESS;
}
*/
